package es.rickyepoderi.jaspi;

import java.io.IOException;
import java.security.Principal;
import java.util.Arrays;
import java.util.Base64;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;
import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.UnsupportedCallbackException;
import javax.security.auth.message.AuthException;
import javax.security.auth.message.AuthStatus;
import javax.security.auth.message.MessageInfo;
import javax.security.auth.message.MessagePolicy;
import javax.security.auth.message.callback.CallerPrincipalCallback;
import javax.security.auth.message.callback.GroupPrincipalCallback;
import javax.security.auth.message.module.ServerAuthModule;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.ws.rs.core.HttpHeaders;

/**
 * <p>Example implementation of a JSPIC ServerAuthModule that authenticates
 * users using BASIC auth and with the information passed in the Map options.
 * The options maps represents the users in the form:
 * "username" => "password###role1,role2,role3"</p>
 *
 * <p>Other examples:</p>
 * <ul>
 * <li>https://github.com/arjantijms/javaee6-auth-example/blob/master/src/jaspic/TestServerAuthModule.java</li>
 * <li>https://github.com/AdamBien/secspike/blob/master/todo/src/main/java/org/omnifaces/security/jaspic/authmodules/BasicAuthModule.java</li>
 * </ul>
 * 
 * <p>How to deploy the module:</p>
 *
 * <pre>
 * module add --name=es.rickyepoderi.jaspi --resources=/path/to/jaspi-module-1.0.jar \
 *   --dependencies=javax.servlet.api,javax.security.auth.message.api
 * </pre>
 *
 * @author rickyepoderi
 */
public class ExampleServerAuthModule implements ServerAuthModule {

    private final Map<String, String> users;
    private final Map<String, Set<String>> roles;
    private CallbackHandler handler;

    public static String ROLES_SEPARATOR = "[\\s]*,[\\s]*";
    public static String PART_SEPARATOR = "###";

    public ExampleServerAuthModule() {
        users = new HashMap<>();
        roles = new HashMap<>();
    }

    /**
     * @inheritDoc
     */
    @Override
    public void initialize(MessagePolicy requestPolicy, MessagePolicy responsePolicy, CallbackHandler handler, Map options) throws AuthException {
        this.handler = handler;
        // loop over options and create the users and roles sets
        if (options != null) {
            for (Map.Entry<Object,Object> e : (Set<Map.Entry<Object,Object>>) options.entrySet()) {
                if (e.getKey() != null && e.getValue() != null) {
                    String username = e.getKey().toString();
                    String value = e.getValue().toString();
                    String[] parts = value.split(PART_SEPARATOR);
                    if (parts.length > 0) {
                        String password = parts[0];
                        if (!username.isEmpty() && !password.isEmpty()) {
                            users.put(username, password);
                            if (parts.length > 1) {
                                String[] rolesArray = parts[1].split(ROLES_SEPARATOR);
                                Set<String> rolesSet = Arrays.stream(rolesArray).filter(r -> !r.isEmpty()).collect(Collectors.toSet());
                                if (!rolesSet.isEmpty()) {
                                    roles.put(username, rolesSet);
                                }
                            }
                        }
                    }
                }
            }
        }
        System.err.println("Initialized users=" + users + " roles=" + roles);
    }

    /**
     * @inheritDoc
     */
    @Override
    public Class[] getSupportedMessageTypes() {
        return new Class[]{HttpServletRequest.class, HttpServletResponse.class};
    }

    private String[] getCredentials(HttpServletRequest request) {
        String header = request.getHeader(HttpHeaders.AUTHORIZATION);
        if (header != null && header.startsWith("Basic ")) {
            String decoded = new String(Base64.getDecoder().decode(header.substring(6)));
            return decoded.split(":");
        } else {
            return null;
        }
    }

    private void applyCallbacks(String username, Subject clientSubject) throws IOException, UnsupportedCallbackException {
        Set<String> userRoles = roles.get(username);
        Callback[] callbacks = new Callback[userRoles == null ? 1 : 2];
        callbacks[0] = new CallerPrincipalCallback(clientSubject, username);
        if (userRoles != null) {
            callbacks[1] = new GroupPrincipalCallback(clientSubject, userRoles.toArray(new String[0]));
        }
        System.err.println("Applying callback for user=" + username + " roles=" + userRoles);
        handler.handle(callbacks);
    }

    /**
     * @inheritDoc
     */
    @Override
    public AuthStatus validateRequest(MessageInfo messageInfo, Subject clientSubject, Subject serviceSubject) throws AuthException {
        try {
            HttpServletRequest request = (HttpServletRequest) messageInfo.getRequestMessage();
            HttpServletResponse response = (HttpServletResponse) messageInfo.getResponseMessage();
            AuthStatus result = AuthStatus.SEND_FAILURE;
            String[] creds = getCredentials(request);
            Principal userPrincipal = request.getUserPrincipal();
            if (userPrincipal != null) {
                // user already logged in, just re-authenticate
                applyCallbacks(userPrincipal.getName(), clientSubject);
                result = AuthStatus.SUCCESS;
            } else if (creds != null && creds.length == 2 && creds[0] != null && creds[1] != null && creds[1].equals(users.get(creds[0]))) {
                // password OK => set the username and roles as principals
                applyCallbacks(creds[0], clientSubject);
                // use jaspi registerSession to cache the principal
                // https://arjan-tijms.omnifaces.org/2013/04/whats-new-in-java-ee-7s-authentication.html
                messageInfo.getMap().put("javax.servlet.http.registerSession", Boolean.TRUE.toString());
                result = AuthStatus.SUCCESS;
            } else {
                // error, just request authentication using 401 and basic
                response.setHeader(HttpHeaders.WWW_AUTHENTICATE, "Basic realm=\"jaspi realm\"");
                response.sendError(HttpServletResponse.SC_UNAUTHORIZED);
            }
            return result;
        } catch (IOException | UnsupportedCallbackException e) {
            throw new AuthException(e.getMessage());
        }
    }

    /**
     * @inheritDoc
     */
    @Override
    public AuthStatus secureResponse(MessageInfo messageInfo, Subject serviceSubject) throws AuthException {
        return AuthStatus.SEND_SUCCESS;
    }

    /**
     * @inheritDoc
     */
    @Override
    public void cleanSubject(MessageInfo messageInfo, Subject subject) throws AuthException {
        if (subject != null) {
            subject.getPrincipals().clear();
            subject.getPrivateCredentials().clear();
            subject.getPublicCredentials().clear();
        }
    }
}